package client;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.UnsupportedEncodingException;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.SocketException;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;

import net.AppServerSocket;
import net.AppSocket;
import net.ByteOp;
import net.ConnectionInfo;
import net.TimeoutLimitException;
import net.Transport;
import server.LoginException;
import server.RepositorioUsers;
import server.UnnavailableLoginException;
import server.User;
import voip.Voice;

public class Client extends Thread{
	ServerConnection server;
	private boolean logado;
	private boolean conectado;
	private boolean vive;
	private FriendThread ft;
	private AppServerSocket requestSocket;
	private int nextId;
	private ConcurrentHashMap<InetAddress, AppSocket> tabelaDeSockets;
	private ConcurrentHashMap<InetAddress, ChatThread> listaDeThreads;
	private Transport trans;
	private ClientGuiListener listener;
	private String login;
	private ArrayBlockingQueue<Byte> ids;
	private ArrayBlockingQueue<Byte> voipConfirm;
	private ArrayBlockingQueue<Byte> respostaArquivo;
	private ChatListener chatListener;
	public Client(ServerConnection server, Transport t) throws IOException,
	ClassNotFoundException{
		conectado = false;
		logado = false;
		vive = true;
		this.server = server;
		ft = new FriendThread(this);
		requestSocket = new AppServerSocket(t,0);
		tabelaDeSockets = new ConcurrentHashMap<InetAddress, AppSocket>();
		listaDeThreads = new ConcurrentHashMap<InetAddress, ChatThread>();
		this.ids = new ArrayBlockingQueue<Byte>(1);
		this.voipConfirm = new ArrayBlockingQueue<Byte>(1);
		this.respostaArquivo = new ArrayBlockingQueue<Byte>(1);
		nextId = 2;
		this.trans = t;
	}
	public ConcurrentHashMap<InetAddress, AppSocket> getTabelaDeSockets(){
		return tabelaDeSockets;
	}
	public ConcurrentHashMap<InetAddress, ChatThread> getListaDeThreads(){
		return listaDeThreads;
	}
	public void addGuiListener(ClientGuiListener listener){
		this.listener = listener;
		ft.addListener(listener);
	}
	public void addChatListener(ChatListener listener){
		this.chatListener = listener;
	}
	public void cadastrar(User usuario) throws IOException, TimeoutLimitException, UnnavailableLoginException, ServerException{
		this.connect();
		server.cadastrar(usuario);
	}
	public void connect() throws TimeoutLimitException{
	
		if(!conectado){

			server.connect();
			
			conectado = true;
		}

	}
	public Hashtable<String, ConnectionInfo> getFriendList() throws TimeoutLimitException, IOException, ClassNotFoundException, ServerException{
		return server.getOnlineFriends();
	}
	public void logar(String login, String senha) throws TimeoutLimitException, LoginException{
		if(!logado){
			if(!conectado){
				this.connect();
			}
			server.logar(login, senha);
			logado = true;
			ft.start();
		}
	}
	public ArrayList<String> getAmigos() throws TimeoutLimitException, IOException, ClassNotFoundException, ServerException{
		return server.getFriends();
	}
	public void adicionarAmigo(String login) throws LoginException, ServerException, TimeoutLimitException, IOException, UnnavailableLoginException{
		server.addFriend(login);
		
	}

	public void logout() throws TimeoutLimitException{
		this.setOffline();
	}
	public void setOffline(){
		System.out.println("set offline");
		this.logado = false;
		this.conectado = false;
		try {
			server.logout();
		} catch (TimeoutLimitException e) {
			
		}
		ft.kill();
		Enumeration<ChatThread> chatThreads = listaDeThreads.elements();
	
		while(chatThreads.hasMoreElements()){
			chatThreads.nextElement().kill();
		}
		
		ft = new FriendThread(this);
		server.newConnection(new AppSocket(trans,1));
	}
	public ClientGuiListener getListener(){
		return this.listener;
	}
	public void run(){
		AppSocket sock = null;
		while(vive){
			ConnectionInfo ci = requestSocket.receiveRequest();
			System.out.println("Recebeu requisio de: "+ci.toString());
			if(!tabelaDeSockets.containsKey(ci.getIp())){
				System.out.println("No estava na tabela");
				sock = new AppSocket(trans, nextId);
				nextId++;
				sock.match(ci);
				tabelaDeSockets.put(ci.getIp(), sock);
				ChatThread thread = new ChatThread(chatListener, this, sock, ci);
				thread.addListener(this.listener);
				listaDeThreads.put(ci.getIp(), thread);
			}else{
				System.out.println("J estava na tabela");
				sock = new AppSocket(trans, nextId);
				nextId++;
				sock.match(ci);
				listaDeThreads.get(ci.getIp()).kill();
				ChatThread thread = new ChatThread(chatListener, this, sock, ci);
				thread.addListener(this.listener);
				listaDeThreads.put(ci.getIp(), thread);
				tabelaDeSockets.put(ci.getIp(), sock);
				
			}
			if(sock != null){
				sock.acceptConnection();
			}
		}
	}
	public void putRespostaVoz(byte resposta){
		this.voipConfirm.offer(resposta);
	}
	public void putRespostaArquivo(byte resposta){
		this.respostaArquivo.offer(resposta);
	}
	public void mandarArquivo(ConnectionInfo ci, String path) throws IOException, TimeoutLimitException, UsuarioNegouException{
		File arq = new File(path);
		FileData dadosArquivo = new FileData(arq.getName(), arq.length());
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream oos = new ObjectOutputStream(baos);
		oos.writeObject(dadosArquivo);
		baos.flush();
		byte[] dados = baos.toByteArray();
		ByteBuffer bf = ByteBuffer.allocate(dados.length+1);
		bf.put((byte)2);
		bf.put(dados);
		bf.flip();
		
		AppSocket sock = this.connectToClient(ci, listener);
		if(sock == null){
			sock = tabelaDeSockets.get(ci.getIp());
		}
		respostaArquivo.clear();
		sock.send(bf.array());
//		Byte resposta = null;
//		try {
//			resposta = respostaArquivo.poll(30000, TimeUnit.MILLISECONDS);
//		} catch (InterruptedException e) {
//			
//		}
//		if(resposta == null || resposta == 0){
//			throw new UsuarioNegouException("Ou o usuario no respondeu a tempo ou ele negou a conversa");
//		}
	}
	public byte chamadaDeVoz(ConnectionInfo ci) throws TimeoutLimitException, UsuarioNegouException{
		byte[] pacote = {1};
		AppSocket sock = this.connectToClient(ci, listener);
		if(sock == null){
			sock = tabelaDeSockets.get(ci.getIp());
		}
		this.voipConfirm.clear();
		sock.send(pacote);
		Byte resposta = null;
		try {
			resposta = voipConfirm.poll(30000, TimeUnit.MILLISECONDS);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		System.out.println(resposta);
		if(resposta == null || resposta == 0){
			throw new UsuarioNegouException("Ou o usuario no respondeu a tempo ou ele negou a conversa");
		}
		return resposta;
	}
	public void responderChamdaDeVoz(ConnectionInfo ci,byte resposta) throws TimeoutLimitException{
		byte[] pacote = {8, resposta};
		AppSocket sock = this.connectToClient(ci, listener);
		if(sock == null){
			sock = tabelaDeSockets.get(ci.getIp());
		}
		sock.send(pacote);
		
	}
	public byte requestId(ConnectionInfo ci, byte idRetorno) throws TimeoutLimitException{
		AppSocket sock = this.connectToClient(ci, listener);
		if(sock == null){
			sock = this.tabelaDeSockets.get(ci.getIp());
		}
		byte[] pacote = {4, idRetorno};
		sock.send(pacote);
		Byte id = null;
		try {
			id = ids.poll(800, TimeUnit.MILLISECONDS);
		} catch (InterruptedException e) {

		}
		if(id == null){
			throw new TimeoutLimitException();
		}
		return id;
	}
	public void putId(Byte id){
		this.ids.offer(id);
	}
	public void sendMsg(String msg, ConnectionInfo ci, byte id) throws TimeoutLimitException{
		byte[] rawMsg = null;
		try {
			rawMsg = msg.getBytes("UTF-16");
		} catch (UnsupportedEncodingException e) {
			rawMsg = msg.getBytes();
			System.out.println("Charset no suportado");
		} 
		AppSocket sock = this.connectToClient(ci, listener);
		ByteBuffer pacote = ByteBuffer.allocate(rawMsg.length+6);
		pacote.put((byte)0);
		pacote.put(id);
		pacote.put(ByteOp.intToByte(rawMsg.length));
		pacote.put(rawMsg);
		
		pacote.flip();
		if(sock == null){
			sock = tabelaDeSockets.get(ci.getIp());
		}
		System.out.println("-------------"+rawMsg.length);
		try{
			sock.send(pacote.array());
		}catch(TimeoutLimitException e){
			listaDeThreads.get(ci.getIp()).kill();
			sock = this.connectToClient(ci, listener);
			if(sock != null){
				sock.send(pacote.array());
			}else{
				System.out.println("NAO TA REMOVENDO DIREITO");
			}
		}
	}
	public void convidaParaGroup(ConnectionInfo ci, ArrayList<ChatId> participantes, byte id) throws IOException, TimeoutLimitException{
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream oos = new ObjectOutputStream(baos);
		oos.writeObject(participantes);
		oos.flush();
		ByteBuffer bf = ByteBuffer.allocate(baos.size()+2);
		bf.put((byte) 6);
		bf.put(id);
		bf.put(baos.toByteArray());
		bf.flip();
		AppSocket sock = this.connectToClient(ci, listener);
		if(sock == null){
			sock = tabelaDeSockets.get(ci.getIp());
		}
//		ChatThread thread = listaDeThreads.get(ci.getIp());
//		thread.addGroupListener(groupListener);
		sock.send(bf.array());
	}
	public void entrarNoChat(ConnectionInfo ci, byte idDestino, byte idOrigem) throws TimeoutLimitException{
		byte[] pacote = {7,idDestino,idOrigem};
		AppSocket sock = this.connectToClient(ci, listener);
		if(sock == null){
			sock = tabelaDeSockets.get(ci.getIp());
		}
		sock.send(pacote);
	}
	public void sendGroup(String msg, Hashtable<ConnectionInfo, String> tabelaDeParticipantes) throws GroupException{
//		byte[] rawMsg = null;
//		String logins = "";
//		try {
//			rawMsg = msg.getBytes("UTF-16");
//		} catch (UnsupportedEncodingException e) {
//			rawMsg = msg.getBytes();
//			System.out.println("Charset no suportado");
//		}
//		ByteBuffer pacote = ByteBuffer.allocate(rawMsg.length+5);
//		pacote.put((byte)4);
//		pacote.put(ByteOp.intToByte(rawMsg.length));
//		pacote.put(rawMsg);
//		pacote.flip();
//		
//		Enumeration<ConnectionInfo> peers = tabelaDeParticipantes.keys();
//		
//		while(peers.hasMoreElements()){
//			ConnectionInfo temp = peers.nextElement();
//			AppSocket sock = tabelaDeSockets.get(temp.getIp());
//			
//			if(sock != null){
//				try{
//					sock.send(pacote.array());
//				}catch(TimeoutLimitException e){
//					logins += " "+tabelaDeParticipantes.get(temp);
//				}
//			}else{
//				logins += " "+tabelaDeParticipantes.get(temp);
//			}
//		}
//		if(!logins.equals("")){
//			throw new GroupException("No foi possvel enviar a mensagem para os seguintes logins:\n"+logins);
//		}
	}
	private synchronized AppSocket connectToClient(ConnectionInfo ci, ClientGuiListener listener) throws TimeoutLimitException{
		AppSocket retorno = null;
		if(!tabelaDeSockets.containsKey(ci.getIp())){
			ci.setId(0);
			AppSocket sock = new AppSocket(trans, nextId);
			sock.connect(ci);
			ChatThread thread = new ChatThread(chatListener, this, sock, ci);
			listaDeThreads.put(ci.getIp(), thread);
			thread.addListener(listener);
			nextId++;
			tabelaDeSockets.put(ci.getIp(), sock);
			retorno = sock;
		}
		return retorno;
	}
}
